from dataclasses import dataclass, field
from typing import Any, Sequence, Tuple, Type

import numpy as np
import napari
from napari.layers import Image, Labels, Layer
from typing_extensions import Annotated
import inspect
from functools import lru_cache

ImageInput = Annotated[Image, {"label": "Image"}]
LayerInput = Annotated[Layer, {"label": "Image or labels"}]
LabelsInput = Annotated[Labels, {"label": "Labels"}]
global_magic_opts = {"auto_call": True}

next_steps_at_the_beginning = [
            "Remove noise",
            "Remove background",
            "Binarize",
            "Label",
            "Combine",
            "Filter",
        ]
next_steps_after_labeling = [
            "Process labels",
            "Combine labels",
            "Measure labels",
            "Measure labeled image",
            "Compare label images",
            "Measurement",
            "Mesh",
        ]
next_steps_after_measuring = [
            "Label filters",
            "Label neighbor filters"
        ]

@dataclass
class Category:
    name: str
    description: str
    inputs: Sequence[Type]
    default_op: str
    default_values : Sequence[float]
    next_step_suggestions: Sequence[str] = field(default_factory=list)
    output: str = "image"  # or labels or dataframe
    # categories
    include: Sequence[str] = field(default_factory=tuple)
    exclude: Sequence[str] = field(default_factory=tuple)
    # visualization
    color_map : str = "gray"
    blending : str = None
    tool_tip : str = ""
    tools_menu : str = None
    auto_call : bool = True


CATEGORIES = {
    "Remove noise": Category(
        name="Remove noise",
        description="Remove noise from images, e.g. by local averaging and blurring.",
        inputs=(ImageInput,),
        default_op="gaussian_blur (clesperanto)",
        default_values=[1, 1, 0],
        next_step_suggestions=next_steps_at_the_beginning,
        include=("filter", "denoise"),
        exclude=("combine",),
        tools_menu="Filtering / noise removal",
    ),
    "Remove background": Category(
        name="Remove background",
        description="Remove background intensity, e.g. caused\nby out-of-focus light or uneven illumination.",
        inputs=(ImageInput,),
        default_op="top_hat_box (clesperanto)",
        default_values=[10, 10, 0],
        next_step_suggestions=[
            "Binarize",
            "Label",
            "Combine",
            "Filter",
        ],
        include=("filter", "background removal"),
        exclude=("combine",),
        tools_menu="Filtering / background removal",
    ),
    "Filter": Category(
        name="Filter",
        description="Filter images, e.g. to adjust gamma or detect edges.",
        inputs=(ImageInput,),
        default_op="gamma_correction (clesperanto)",
        default_values=[1, 1, 0],
        next_step_suggestions=[
            "Binarize",
            "Label",
            "Combine",
        ],
        include=("filter",),
        exclude=("combine", "denoise", "background removal", "binary processing"),
        tools_menu="Filtering",
    ),
    "Combine": Category(
        name="Combine",
        description="Combine images using pixel-wise mathematical operations.",
        inputs=(LayerInput, LayerInput),
        default_op="add_images (clesperanto)",
        include=("combine",),
        exclude=("map", 'combine labels',),
        default_values=[1, 1],
        next_step_suggestions=[
            "Binarize",
            "Label",
        ],
        tools_menu="Image math",
    ),
    "Transform": Category(
        name="Transform",
        description="Apply spatial transformation to images.",
        inputs=(LayerInput,),
        default_op="sub_stack (clesperanto)",
        output="image",  # can also be labels
        default_values=[0, 0, 0, 1, 1],
        next_step_suggestions=next_steps_at_the_beginning,
        include=("transform",),
        exclude=("combine",),
        tools_menu="Transform",
    ),
    "Projection": Category(
        name="Projection",
        description="Reduce dimensionality of images\nfrom three to two dimensions.",
        inputs=(LayerInput,),
        default_op="maximum_z_projection (clesperanto)",
        default_values=[1, 1, 1],
        next_step_suggestions=next_steps_at_the_beginning,
        output="image",  # can also be labels
        include=("projection",),
        tools_menu="Projection",
    ),
    "Binarize": Category(
        name="Binarize",
        description="Turn images into binary images.",
        inputs=(LayerInput,),
        default_op="threshold_otsu (clesperanto)",
        output="labels",
        default_values=[1, 1, 0],
        next_step_suggestions=[
            "Label",
            "Process labels",
        ],
        include=("binarize",),
        exclude=("combine",),
        tools_menu="Segmentation / binarization",
    ),
    "Label": Category(
        name="Label",
        description="Turn images into label images by labeling objects.",
        inputs=(LayerInput,),
        default_op="voronoi_otsu_labeling (clesperanto)",
        output="labels",
        default_values=[2, 2],
        next_step_suggestions=next_steps_after_labeling,
        include=("label",),
        tools_menu="Segmentation / labeling",
    ),
    "Process labels": Category(
        name="Process labels",
        description="Process label images to improve\nby changing their shape and/or removing\nobjects which don't fulfill certain conditions.",
        inputs=(LabelsInput,),
        default_op="exclude_labels_on_edges (clesperanto)",
        output="labels",
        default_values=[2, 100],
        next_step_suggestions=next_steps_after_labeling,
        include=("label processing",),
        exclude=("combine",),
        tools_menu="Segmentation post-processing",
    ),
    "Combine labels": Category(
        name="Combine labels",
        description="Process label images multiple label image\nto create a new label image.",
        inputs=(LabelsInput,LabelsInput,),
        default_op="combine_labels (clesperanto)",
        output="labels",
        default_values=[2, 100],
        next_step_suggestions=next_steps_after_labeling,
        include=("label processing","combine labels"),
        exclude=(),
        tools_menu="Segmentation post-processing",
    ),
    "Measure labels": Category(
        name="Measure labels",
        description="Measure and visualize spatial\nfeatures of labeled objects.",
        inputs=(LabelsInput,),
        default_op="pixel_count_map (clesperanto)",
        default_values=[1, 1],
        next_step_suggestions=next_steps_after_measuring,
        include=("label measurement", "map"),
        exclude=("combine",),
        color_map="turbo",
        tools_menu="None",
    ),
    "Measure labeled image": Category(
        name="Measure labeled image",
        description="Measure and visualize intensity-based\nfeatures of labeled objects.",
        inputs=(ImageInput, LabelsInput),
        default_op="mean_intensity_map (clesperanto)",
        default_values=[1, 1],
        next_step_suggestions=next_steps_after_measuring,
        include=("combine","label measurement", "map",),
        exclude=("label comparison",),
        color_map="turbo",
        tools_menu="None",
    ),
    "Compare label images": Category(
        name="Compare label images",
        description="Measure and visualize differences \nof labeled objects in two label images.",
        inputs=(LabelsInput, LabelsInput),
        output="image",
        default_op="label_overlap_count_map (clesperanto)",
        default_values=[],
        next_step_suggestions=next_steps_after_measuring,
        include=("combine","label measurement", "map", "label comparison",),
        exclude=(),
        color_map="turbo",
        tools_menu="None",
    ),
    "Label neighbor filters": Category(
        name="Label neighbor filters",
        description="Process values associated with labeled objects\naccording to the neighborhood-graph of the labels.",
        inputs=(ImageInput, LabelsInput),
        default_op="mean_of_n_nearest_neighbors_map (clesperanto)",
        default_values=[1, 100],
        next_step_suggestions=next_steps_after_labeling,
        include=("neighbor",),
        color_map="turbo",
        tools_menu="Label neighbor filters",
    ),
    "Label filters": Category(
        name="Label filters",
        description="Process label images depending on values in corresponding images.\nPleease use parametric maps only as input image.",
        inputs=(ImageInput, LabelsInput),
        default_op="exclude_labels_with_map_values_out_of_range (clesperanto)",
        output="labels",
        default_values=[1, 100],
        next_step_suggestions=next_steps_after_labeling,
        include=('label processing', 'combine'),
        exclude=("neighbor",),
        tools_menu="Segmentation post-processing",
    ),
    "Mesh": Category(
        name="Mesh",
        description="Draw connectivity meshes between\ncentroids of labeled objects.",
        inputs=(LabelsInput,),
        default_op="draw_mesh_between_touching_labels (clesperanto)",
        default_values=[1],
        next_step_suggestions=[],
        include=("label measurement", "mesh"),
        color_map="green",
        blending="additive",
        tools_menu="Visualization",
    ),
    "Measurement": Category(
        name="Measurement",
        description="Measure features and show results in a table.",
        inputs=(ImageInput, LabelsInput),
        output="dataframe",
        default_op="Regionprops (nsr)",
        default_values=[],
        next_step_suggestions=[],
        include=(),
        tools_menu="Measurement",
        auto_call=False
    ),
}



def attach_tooltips():
    """
    Attach tooltips to categories which contain all functions
    This is necessary so that the search later finds operations by name in categories.
    The search searches in the tooltip.

    Todo: This only works if the tools-menu is initialized.
          We should alternatively search for menu names + functions in npe2.
    """
    # attach tooltips
    for k, c in CATEGORIES.items():
        choices = operations_in_menu(c)
        c.tool_tip = c.description + "\n\nOperations:\n* " + "\n* ".join(choices).replace("_", " ")


@lru_cache(maxsize=1)
def all_operations():
    """Get a dictionary of all compatible functions of installed plugins

    Returns
    -------
    dict(str:callable)
        Dictionary of name-function pairs

    """
    # harvest functions from clesperanto
    cle_ops = collect_from_pyclesperanto_if_installed()

    # harvest functions from napari-tools-menu
    tools_ops = collect_from_tools_menu_if_installed()

    # harvest from napari-plugin-enginge-2
    npe2_ops = collect_from_npe2_if_installed()

    # combine all
    all_ops = {**cle_ops, **tools_ops, **npe2_ops}
    return all_ops


def get_name_of_function(func):
    """
    Searches all functions for a given function
    and returns its human-readable name
    """
    for k, v in all_operations().items():
        if v is func or v.__wrapped__ is func:
            if ">" in k:
                return k.split(">")[1]
            else:
                return k
    return None


def collect_from_pyclesperanto_if_installed():
    """
    Collect all functions from clesperanto that are annotated with "in assistant"
    """
    from napari_time_slicer import time_slicer
    try:
        import pyclesperanto_prototype as cle
    except ImportError:
        print("Assistant skips harvesting pyclesperanto as it's not installed.")
        return {}

    result = {}

    for k, c in CATEGORIES.items():
        if not callable(c):
            choices = cle.operations(['in assistant'] + list(c.include), c.exclude)
            for choice, func in choices.items():
                result[c.tools_menu + ">" + choice + " (clesperanto)"] = time_slicer(func)

    return result


def collect_from_tools_menu_if_installed():
    """
    Collect all functions that process images (no dock-widgets and actions) from napari-tools-menu
    """
    try:
        from napari_tools_menu import ToolsMenu
    except ImportError:
        print("Assistant skips harvesting tools menu as it's not installed.")
        return {}

    import pandas

    allowed_types = ["napari.types.LabelsData", "napari.types.ImageData", "int", "float", "str", "bool",
                     "napari.viewer.Viewer", "napari.Viewer", "magicgui.types.PathLike", "typing.Union[pathlib.Path, str, bytes]"]
    allowed_types = allowed_types + ["<class '" + t + "'>" for t in allowed_types]

    result = {}
    for k, v in ToolsMenu.menus.items():
        typ = v[1]
        if typ == "function":
            f = v[0]
            sig = inspect.signature(f)

            # all parameters must be images, labels, int, float or str
            skip = False
            for i, key in enumerate(list(sig.parameters.keys())):
                type_annotation = str(sig.parameters[key].annotation)
                if not "function NewType.<locals>.new_type" in type_annotation:
                    if type_annotation not in allowed_types:
                        # print("Skip", k, "because", str(type_annotation), "not in allowed types")
                        skip = True
                        break
            if skip:
                continue

            # return type must be image or label_image
            if sig.return_annotation not in [napari.types.LabelsData, "napari.types.LabelsData",
                                             napari.types.ImageData, "napari.types.ImageData",
                                             pandas.DataFrame, "pandas.DataFrame"]:
                continue

            result[k] = f

    return result


def collect_from_npe2_if_installed():
    """
    Collect all functions provided by the NPE2 interface (napari) which contain
    a menu name
    """
    try:
        import npe2
    except ImportError:
        print("Assistant skips harvesting npe2 as it's not installed.")
        return {}
    pm = npe2.PluginManager.instance()

    result = {}
    for pname, item in pm._manifests.items():
        if item.contributions.widgets is not None:
            for c in item.contributions.widgets:
                wname = c.display_name
                # only harvest items which look like a menu decription
                # todo: as soon as npe2 supports menus, change this here
                if ">" in pname or ">" in wname:
                    # print("loading ", pname, wname)

                    t = get_widget_contribution(pname, wname)
                    if t is not None:
                        factory = t[0]
                        menu = t[1]
                        kwargs = {}
                        w = factory(**kwargs)
                        menu = menu.replace(" > ", ">")
                        result[menu] = w._function
    return result


# source: https://github.com/napari/napari/blob/1363bd47da668a5826a75a73be93f7a7f8042fd7/napari/plugins/_npe2.py#L91
def get_widget_contribution(
        plugin_name: str, widget_name: str = None
):
    import npe2
    for contrib in npe2.PluginManager.instance().iter_widgets():
        if contrib.plugin_name == plugin_name and (
                not widget_name or contrib.display_name == widget_name
        ):
            return contrib.get_callable(), contrib.display_name
    return None


def filter_operations(menu_name: str):
    """
    Find functions that contain a given name
    Parameters
    ----------
    menu_name: str
        case sensitive search string

    Returns
    -------
    dict(str:callable)
        Dictionary of name-function pairs
    """
    result = {}
    for k,v in all_operations().items():
        if menu_name+">" in k:
            result[k] = v
    return result

def operations_in_menu(category, search_string: str = None):
    """
    Return all functions as list in a given category that contain a
    given search string.
    """
    if not hasattr(category, "tools_menu"):
        return []
    menu_name = category.tools_menu
    choices = filter_operations(menu_name)

    all_ops = all_operations()

    if search_string is not None and len(search_string) > 0:
        choices_dict = {k: all_ops[k] for k in choices}
        choices = [c for c, v in choices_dict.items() if search_string in c.lower() or (v.__doc__ is not None and search_string in v.__doc__.lower())]
    choices = [c.split(">")[1].strip() for c in choices]
    choices = sorted(choices, key=str.casefold)

    # check if the image parameters fit
    result = []
    for name in choices:
        func = find_function(name)
        sig = inspect.signature(func)

        # count number of image-like parameters and compare to category
        num_image_parameters_in_category = len(category.inputs)
        num_image_parameters_in_function = 0
        for i, key in enumerate(list(sig.parameters.keys())):
            type_annotation = str(sig.parameters[key].annotation)

            if "NewType.<locals>.new_type" in type_annotation or \
                "Image" in type_annotation or \
                "LabelsData" in type_annotation or \
                "LayerData" in type_annotation:
                num_image_parameters_in_function = num_image_parameters_in_function + 1
            else:
                break

        if "pyclesperanto_prototype" in func.__module__:
            # all clesperanto function have an output image which we don't pass
            num_image_parameters_in_function -= 1

        # only keep the function in this category if it matches
        if num_image_parameters_in_category == num_image_parameters_in_function:
            if category.name == "Measurement":
                if str(sig.return_annotation) == "pandas.DataFrame":
                    result.append(name)
            else:
                result.append(name)

    return result


def find_function(op_name):
    """
    Find a function by name (in menu)
    """
    all_ops = all_operations()
    found_function = None

    # find exact match first
    for k, f in all_ops.items():
        if ">" in k:
            k = k.split(">")[1]
        if op_name == k:
            found_function = f
            break

    if found_function is None:
        # find approximate match
        for k, f in all_ops.items():
            if op_name in k:
                found_function = f

    if found_function is None:
        print("No function found for", op_name)
    return found_function


def get_category_of_function(func = None, func_name = None):
    """
    Searches categories for a given function and returns the first (!)
    category that contains the function
    """
    
    if func_name is None:
        func_name = get_name_of_function(func)

    for k, c in CATEGORIES.items():
        if not callable(c):
            ops = operations_in_menu(c)
            if func_name in ops:
                return c
    return None


def filter_categories(search_string: str = ""):
    """Return all categories that have a function that contains a
    given search string in their name.
    """
    if search_string is None or len(search_string) == 0:
        search_string = ""

    from copy import copy

    all_categories = {}
    for k, c in CATEGORIES.items():
        #if callable(c) or search_string in c.tool_tip.lower():
        new_c = copy(c)
        all_categories[k] = new_c

    category_found = False
    result = {}
    for k, c in all_categories.items():
        if callable(c):
            if category_found ^ ("Search" in k):
                result[k] = c
        else:
            choices = operations_in_menu(c, search_string)
            c.tool_tip = c.description + "\n\nOperations:\n* " + "\n* ".join(choices).replace("_", " ")
            if len(choices) > 0:
                result[k] = c
                category_found = True

    return result



def main():
    import sys
    from ._viewer import Viewer
    from qtpy.QtCore import QTimer

    viewer = Viewer()

    if len(sys.argv) > 1:
        def later():
            for pos, arg in enumerate(sys.argv):
                print('Argument %d: %s' % (pos, arg))
                if pos > 0:
                    viewer.open(arg)

        QTimer.singleShot(600, later)

    import napari
    napari.run()


if __name__ == '__main__':
    main()


from inspect import signature
from napari.utils._magicgui import _make_choice_data_setter
from ._categories import get_category_of_function, get_name_of_function
from ._gui._category_widget import (
    separate_argnames_by_type,
    kwarg_key_adapter,
    make_gui_for_category,
    DEFAULT_BUTTON_SIZE,
    LAYER_NAME_PREFIX,
)
from napari import Viewer
from napari_workflows import Workflow

# TODO look if comments are accurate
def initialise_root_functions(workflow: Workflow, viewer: Viewer, button_size: int = DEFAULT_BUTTON_SIZE):
    """
    Makes widgets for all functions which have a root image as input. The widgets are 
    added to the viewer and correct input images must be chosen to complete the loading
    of the workflow if they are not already present in the layers

    Parameters
    ----------
    workflow:
        napari-workflow object
    viewer:
        napari.Viewer instance
    button_size: int
        button size parameter that is passed to make_category_gui    
    """
    widget_dw = []

    # find all workflow steps with functions which have root images as an input
    root_functions = wf_steps_with_root_as_input(workflow)
    layer_names = [lay.name for lay in viewer.layers]

    # iterate over root function workflow steps
    for wf_step_name in root_functions:
        # get the fuction from the workflow step and change its signature to
        # the values specified by the workflow
        func = workflow._tasks[wf_step_name][0]

        # create the widget based on the adjusted function and turn off autocall
        # for functions with more than 1 input image
        sources = workflow.sources_of(wf_step_name)
        
        # if more than one source is needed autocall needs to be set to false
        # in order to avoid crashing
        if len(sources) > 1:
            widget = make_flexible_gui(
                func, 
                viewer,
                autocall= False,
                button_size = button_size,
            )
        else:
            widget = make_flexible_gui(
                func, 
                viewer, 
                button_size = button_size,
            )
            widget._auto_call = False

        # determine if all input images are in layer names
        sources_present = True
        for source in sources:
            if source not in layer_names:
                sources_present = False

        # if all input images are present in the layers we can preselct them
        if sources_present:
            dw = viewer.window.add_dock_widget(widget, name= wf_step_name[10:])
            set_choices(workflow= workflow,
                        wf_step_name= wf_step_name,
                        viewer= viewer,
                        widget= widget)
        else:
            # if the input images aren't present we will leave the dropdowns
            # open to chose the right image - make a tooltip which tells the 
            # user to select the input image specified by the workflow
            key_source_list = get_source_keywords_and_sources(workflow,
                                                            wf_step_name)
            for key, source in key_source_list:
                widget[key].tooltip = f'Select {source} or equivalent'

            # add the final widget to the napari viewer
            dw = viewer.window.add_dock_widget(widget, name = wf_step_name[10:] + '<b> - SELECT INPUT</b>')
        
        # setting the right parameters
        cat_kwargs = category_kwargs(
            func,
            kwargs_of_wf_step(workflow,wf_step_name),
        )
        for k,v in cat_kwargs.items():
            widget[k].value = v
        
        if len(sources) <2:
            widget._auto_call = True
        # calling the widget with the correct input images
        widget()
        widget_dw.append((widget,dw))
    
    return widget_dw

def load_remaining_workflow(workflow: Workflow, viewer: Viewer, button_size: int = DEFAULT_BUTTON_SIZE):
    """
    Loads the remaining workflow once initialise_root_functions has been called with
    the same workflow and the same napari viewer with given button sie. Returns widgets
    and respective dockwidgets as a list of tupules: [(widget, dockwidget)]

    Parameters
    ----------
    workflow:
        napari-workflow object
    viewer:
        napari.Viewer instance
    button_size: int
        button size parameter that is passed to make_category_gui
    """
    # find all workflow steps with functions which have root images as an input
    root_functions = wf_steps_with_root_as_input(workflow)
    # get the layer object from the napari viewer
    layers = viewer.layers

    # start the iteration with the followers of the root functions
    followers = []
    widget_dw = []
    for root in root_functions:
        followers += workflow.followers_of(root)

    # iteration over followers: the list of followers gets bigger the more functions 
    # have been added with followers of their own
    for i,follower in enumerate(followers):
        # get current layer names
        layer_names = [str(lay) for lay in layers]
        # find all sources of the current function being added
        sources = workflow.sources_of(follower)

        # checking if we have all input images to the function in the napari layers
        sources_present = True
        for source in sources:
            if source not in layer_names:
                sources_present = False

        # if some input images are missing we will move to other functions first
        if not sources_present:
            if follower not in followers[i+1:]:
                followers.append(follower)
            continue    

        # if all input images are there we can continue
        
        # get the fuction from the workflow step
        func = workflow._tasks[follower][0]

        # if more than one source is needed autocall needs to be set to false
        # in order to avoid crashing
        if len(sources) > 1:
            widget = make_flexible_gui(func, 
                                        viewer,
                                        autocall= False,
                                        button_size = button_size
            )
        else:
            widget = make_flexible_gui(func, 
                                        viewer, 
                                        button_size = button_size
            )
            widget._auto_call = False
        # add the final widget to the napari viewer and set the input images in
        # the dropdown to the specified input images
        dw = viewer.window.add_dock_widget(
            widget, 
            name= follower[len(LAYER_NAME_PREFIX):],
            )
        set_choices(workflow= workflow,
                    wf_step_name= follower,
                    viewer= viewer,
                    widget= widget)
        
        # setting the right parameters
        cat_kwargs = category_kwargs(
            func,
            kwargs_of_wf_step(workflow,follower),
        )
        for k,v in cat_kwargs.items():
            widget[k].value = v

        if len(sources) <2:
            widget._auto_call = True
        # calling the widget with the correct input images
        widget()

        widget_dw.append((widget,dw))
        # finding new followers of the current workflow step
        new_followers = workflow.followers_of(follower)

        # checking if the new followers are already in the que of workflow steps
        # to be added
        for new_follower in new_followers:
            if new_follower not in followers[i+1:]:
                followers.append(new_follower)
        
    return widget_dw

def make_flexible_gui(func, viewer: Viewer, autocall: bool = True, button_size: int = DEFAULT_BUTTON_SIZE):
    """
    Function returns a widget with a GUI for the function provided in the parameters,
    that can be added to the napari viewer.

    Parameters
    ----------
    func: 
        input function to generate the gui for
    viewer:
        napari.Viewer instance to which the widget is added
    wf_step_name: str
        name of the workflow step matching the function
    autocall: Boolean
        sets the auto_call behaviour of the magicgui.magicgui function
    button_size: int
        button size parameter that is passed to make_category_gui
    """
    gui = None
    category = get_category_of_function(func)

    if category is None:
        raise ModuleNotFoundError("Cannot build user interface for not installed function " + str(func))
    else:
        gui = make_gui_for_category(
            category, 
            viewer=viewer, 
            operation_name=get_name_of_function(func), 
            autocall=autocall, 
            button_size= button_size
        )

    return gui

def set_choices(workflow: Workflow, wf_step_name: str, viewer: Viewer, widget):
    """
    Sets the choices for image drop down menu to the images specified by the workflow

    Parameters
    ----------
    workflow:
        napari-workflow object
    wf_step_name: str
        the string of the workflow step for which the choices in the widget will
        be modified
    viewer:
        napari.Viewer instance
    widget:
        the magicgui FunctionGui object for which the input choices should be changed
    """
    func = workflow._tasks[wf_step_name][0]
    args = workflow._tasks[wf_step_name][1:]
    sources = workflow.sources_of(wf_step_name)
    adapter = kwarg_key_adapter(func)
    keyword_list = list(signature(func).parameters.keys())
    image_keywords = [(key,value) for key, value in zip(keyword_list,args) if value in sources]

    for i, (key, name) in enumerate(image_keywords):
        widget[adapter[key]].choices = get_layers_data_of_name(name, viewer, widget[adapter[key]])

def get_layers_data_of_name(layer_name: str, viewer: Viewer, gui):
    """
    Returns a choices dictionary which can be utilised to set an input image of a 
    widget with the function set_choices. code modified from napari/utils/_magicgui
    get_layers_data function.

    Parameters
    ----------
    layer_name: str
        the layer which should be selected as the only choice in a drop down menu
    viewer:
        napari.Viewer instance
    gui:
        magicgui Combobox instance for which the choices should
    """
    choices = []
    for layer in [x for x in viewer.layers if str(x) == layer_name]:
        choice_key = f'{layer.name}'
        choices.append((choice_key, layer.data))
        layer.events.data.connect(_make_choice_data_setter(gui, choice_key))

    return choices

def wf_steps_with_root_as_input(workflow: Workflow):
    """
    Returns a list of workflow steps that have root images as an input

    Parameters
    ----------
    workflow: 
        napari_workflows Workflow class
    """
    roots = workflow.roots()
    wf_step_with_rootinput = []
    for result, task in workflow._tasks.items():
            for source in task:
                if isinstance(source, str):
                    if source in roots:
                        wf_step_with_rootinput.append(result)
                        break
    return wf_step_with_rootinput

def get_source_keywords_and_sources(workflow: Workflow, wf_step_name: str):
    """
    Returns a list of tuples containing (function_keyword, image_name) for all 
    sources of a workflow step with name: wf_step_name

    Parameters
    ----------
    workflow: 
        napari_workflows Workflow class
    wf_step_name: str
        name of the workflow step for which function keywords and image names
        should be returned
    """
    func = workflow._tasks[wf_step_name][0]
    args = workflow._tasks[wf_step_name][1:]
    adapter = kwarg_key_adapter(func)
    
    sources = workflow.sources_of(wf_step_name)
    keyword_list = list(signature(func).parameters.keys())
    image_keywords = [
        (adapter[key],value) for key, value in zip(keyword_list,args) if value in sources
    ]
    
    return image_keywords

def category_kwargs(func,kwargs):
    """
    Returns a dict of kwargs mapping the napari assistant keywords to 
    the values specified by the functions

    Parameters
    ----------
    func: function
        input function for which the assistant kwargs should be generated for
    kwargs: dict
        kwargs of func
    """
    category_kwargs = {}
    adapter = kwarg_key_adapter(func)

    sig = signature(func)
    # get the names of positional parameters in the new operation
    param_names, foo, bar, foobar, fbar = separate_argnames_by_type(
        sig.parameters.items())

    category_kwargs = {adapter[key] : kwargs[key] for key in param_names}

    return category_kwargs

def kwargs_of_wf_step(
    workflow: Workflow, 
    wf_step_name: str
) -> dict:
    """
    Returns a dictionary with kwargs specified by the function and parameters
    of a specific workflow step

    Parameters
    ----------
    workflow:
        workflow object specifying functions and parameters
    wf_step_name:
        name of the worflow step for which the kwargs will be generated
    """
    func     = workflow._tasks[wf_step_name][0]
    arg_vals = workflow._tasks[wf_step_name][1:]

    # getting the keywords corresponding to the values
    keyword_list = list(signature(func).parameters.keys())

    # creating the kwargs dict
    kw_dict = {}
    for kw, val in zip(keyword_list, arg_vals):
        kw_dict[kw] = val
        
    return kw_dict




def Viewer(show_assistant: bool = True, *args, **kwargs):
    import napari
    from qtpy.QtCore import QTimer

    viewer = napari.Viewer(*args, **kwargs)

    def later():
        if show_assistant:
            from ._gui._Assistant import Assistant
            viewer.window.add_dock_widget(Assistant(viewer), name="Assistant")

    QTimer.singleShot(300, later)

    return viewer



from __future__ import annotations

from inspect import Parameter, Signature, signature

from qtpy import QtCore
from qtpy.QtCore import QTimer
from qtpy.QtWidgets import QDoubleSpinBox, QComboBox, QWidget

from typing import Any, Optional, TYPE_CHECKING, Sequence
from functools import partial

import toolz
from loguru import logger
from magicgui import magicgui
from typing_extensions import Annotated
import napari
import numpy as np

from .._categories import Category, find_function, get_name_of_function
from qtpy.QtWidgets import QPushButton, QDockWidget
from magicgui.types import PathLike

try:
    import pyclesperanto_prototype as cle
    Image_type = cle.Image
except:
    Image_type = np.ndarray

if TYPE_CHECKING:
    from napari.layers import Layer
    from napari import Viewer

VIEWER_PARAM = "viewer"
OP_NAME_PARAM = "op_name"
OP_ID = "op_id"
DEFAULT_BUTTON_SIZE = 40
LAYER_NAME_PREFIX = "Result of "
# We currently support operations with up to 6 numeric parameters, 3 booleans and 3 strings (see lists below)
FloatRange = Annotated[float, {"min": np.finfo(np.float32).min, "max": np.finfo(np.float32).max, "step": 1}]
BoolType = Annotated[bool, {}]
StringType = Annotated[str, {}]
PathLikeType = Annotated[PathLike, {}]

PositiveFloatRange = Annotated[float, {"min": 0, "max": np.finfo(np.float32).max}]
category_args = [
    ("x", FloatRange, 0),
    ("y", FloatRange, 0),
    ("z", FloatRange, 0),
    ("u", FloatRange, 0),
    ("v", FloatRange, 0),
    ("w", FloatRange, 0),
    ("a", BoolType, False),
    ("b", BoolType, False),
    ("c", BoolType, False),
    ("d", BoolType, False),
    ("e", BoolType, False),
    ("f", BoolType, False),
    ("g", BoolType, False),
    ("h", BoolType, False),
    ("i", BoolType, False),
    ("j", BoolType, False),
    ("k", StringType, ""),
    ("l", StringType, ""),
    ("m", StringType, ""),
    ("o", PathLikeType, ""),
    ("p", PathLikeType, ""),
    ("q", PathLikeType, ""),
]
category_args_numeric = ["x", "y", "z", "u", "v", "w"]
category_args_bool = ["a", "b", "c", "d", "e", "f", "g","h","i","j"]
category_args_text = ["k", "l", "m"]
category_args_file = ["o", "p", "q"]

def num_positional_args(func, types=[np.ndarray, napari.types.ImageData, napari.types.LabelsData, Image_type, int, str, float, bool]) -> int:
    params = signature(func).parameters
    return len([p for p in params.values() if p.annotation in types])

def kwarg_key_adapter(func):
    '''
    Returns a dictionary mapping keywords between the category kwarg keys used
    by the napari assistant to the kwarg keys of the input functions as well
    as the reverse mapping.

    Parameters
    ----------

    func: function
        the function for which the mapping should be generated
    '''
    adapter = {}
    sig = signature(func)
    items = sig.parameters.items()
    # get the names of positional parameters in the new operation
    param_names, numeric_param_names, bool_param_names, str_param_names, file_param_names = separate_argnames_by_type(
        items)

    # go through all parameters and collect their values in an args-array
    num_count = 0
    str_count = 0
    bool_count = 0
    file_count = 0
    for key in param_names: 
        if key in numeric_param_names:
            adapter[category_args_numeric[num_count]] = key
            adapter[key] = category_args_numeric[num_count]
            num_count += 1
        elif key in bool_param_names:
            adapter[category_args_bool[bool_count]] = key
            adapter[key] = category_args_bool[bool_count]
            bool_count += 1
        elif key in str_param_names:
            adapter[category_args_text[str_count]] = key
            adapter[key] = category_args_text[str_count]
            str_count += 1
        elif key in file_param_names:
            adapter[category_args_file[file_count]] = key
            adapter[key] = category_args_file[file_count]
            file_count += 1

    other_count = 0    
    other_param_names = [
        name
        for name, param in items
        if param.annotation not in {int, str, float, bool, PathLike}
    ]
    for key in other_param_names:
        adapter["input" + str(other_count)] = key
        adapter[key] = "input" + str(other_count)
        other_count += 1

    return adapter

@logger.catch
def call_op(op_name: str, inputs: Sequence[Layer], timepoint : int = None, viewer: napari.Viewer = None, **kwargs) -> np.ndarray:
    """Call operation `op_name` with specified inputs and args.

    Takes care of transfering data to GPU and omitting extra positional args

    Parameters
    ----------
    op_name : str
        name of operation to execute.
    inputs : Sequence[Layer]
        The napari layer inputs

    Returns
    -------
    np.ndarray
        result image
    """

    if not inputs or inputs[0] is None:
        return

    i0 = inputs[0].data
    gpu_ins = [i.data if i is not None else i0 for i in inputs]

    # call actual cle function ignoring extra positional args
    cle_function = find_function(op_name)
    nargs = num_positional_args(cle_function)

    new_sig = signature(cle_function)
    adapter = kwarg_key_adapter(cle_function)
    # get the names of positional parameters in the new operation
    param_names, _, _, _, _ = separate_argnames_by_type(
        new_sig.parameters.items())
    
    args = tuple([kwargs[adapter[key]] for key in param_names])

    if cle_function.__module__ == "pyclesperanto_prototype":
        # todo: we should handle all functions equally
        gpu_out = None

        logger.info(f"{op_name}(..., {', '.join(map(str, args))})")
        args = ((*gpu_ins, gpu_out) + args)[:nargs]

        gpu_out = cle_function(*args, viewer=viewer)

        # return output
        return gpu_out, args
    else:
        args = (*gpu_ins, *args)[:nargs+1]

        import inspect
        sig = inspect.signature(cle_function)

        # pass viewer if requested
        kwargs = {}
        for k, v in sig.parameters.items():
            if k == "viewer" or k == "napari_viewer" or "napari.viewer.Viewer" in str(v):
                kwargs[k] = viewer

        # Make sure that the annotated types are really passed to a given function
        for i, k in enumerate(list(sig.parameters.keys())):
            if i >= len(args):
                break
            type_annotation = str(sig.parameters[k].annotation)
            #print("Annotation:", type_annotation)
            args = list(args)
            for typ in ["int", "float", "str"]:
                if typ in type_annotation:
                    converter = eval(typ)
                    #print("converter", converter)
                    args[i] = converter(args[i])

        gpu_out = cle_function(*args, **kwargs)

        if sig.return_annotation in [napari.types.LabelsData, "napari.types.LabelsData"]:
            if gpu_out.dtype is not int:
                gpu_out = gpu_out.astype(int)

        return gpu_out, args


def _show_result(
    gpu_out: np.ndarray,
    viewer: Viewer,
    name: str,
    layer_type: str,
    op_id: int,
    translate=None,
    cmap=None,
    blending=None,
    scale=None,
) -> Optional[Layer]:
    """Show `gpu_out` in the napari viewer.

    Parameters
    ----------
    gpu_out : np.ndarray
        an image to show
    viewer : napari.Viewer
        The napari viewer instance
    name : str
        The name of the layer to create or update.
    layer_type : str
        the layer type to create (must be 'labels' or 'image)
    op_id : int
        an ID to associate with the newly created layer (will be added to
        layer.metada['op_id'])
    translate : [type], optional
        translate parameter for layer creation, by default None
    cmap : str, optional
        a colormap to use for images, by default None
    blending : str, optional
        blending mode for visualization, by default None

    Returns
    -------
    layer : Optional[Layer]
        The created/udpated layer, or None if no viewer is present.
    """
    if layer_type == "dataframe":
        # todo: in case an operation returns a dataframe; do we need to do anything with it?
        return None

    #print("OP ID ", op_id)
    if not viewer:
        logger.warning("no viewer, cannot add image")
        return
    # show result in napari
    if "dask" in str(type(gpu_out)):
        clims = None
    else:
        clims = [gpu_out.min(), gpu_out.max()]

        if clims[1] == 0:
            clims[1] = 1

        if clims[0] == clims[1]:
            clims = None

    # conversion will be done inside napari. We can continue working with the potentially OCL-array from here.
    data = gpu_out

    try:
        # look for an existing layer
        layer = next(x for x in viewer.layers if isinstance(x.metadata, dict) and x.metadata.get(OP_ID) == op_id)
        # logger.debug(f"updating existing layer: {layer}, with id: {op_id}")
        layer.data = data
        layer.name = name
        # layer.translate = translate
    except StopIteration:
        # otherwise create a new one
        # logger.debug(f"creating new layer for id: {op_id}")
        add_layer = getattr(viewer, f"add_{layer_type}")
        kwargs = dict(name=name, metadata={OP_ID: op_id})
        if layer_type == "image":
            kwargs["colormap"] = cmap
            kwargs["blending"] = blending
            kwargs['contrast_limits'] = clims

        layer = add_layer(data, **kwargs)

    if scale is not None:
        if len(layer.data.shape) == len(scale):
            layer.scale = scale
        if len(layer.data.shape) < len(scale):
            layer.scale = scale[-len(layer.data.shape):]
    return layer


def _generate_signature_for_category(category: Category, search_string:str= None, viewer:napari.Viewer = None) -> Signature:
    """Create an inspect.Signature object representing a cle Category.

    The output of this function can be used to set function.__signature__ so that
    magicgui can convert it into the appropriate widget.
    """
    if search_string is not None:
        search_string = search_string.lower()

    k = Parameter.KEYWORD_ONLY

    # add inputs (we name them inputN ...)
    params = [
        Parameter(f"input{n}", k, annotation=t) for n, t in enumerate(category.inputs)
    ]
    # Add valid operations choices (will create the combo box)
    from .._categories import operations_in_menu
    choices = list(operations_in_menu(category, search_string))
    #print("choices:", choices)
    op_type = Annotated[str, {"choices": choices, "label": "Operation"}]
    default_op = category.default_op
    if not any(default_op == op for op in choices):
        #print("Default-operation is not in list!")
        default_op = None

    if default_op is None:
        params.append(
            Parameter(OP_NAME_PARAM, k, annotation=op_type)
        )
    else:
        params.append(
            Parameter(OP_NAME_PARAM, k, annotation=op_type, default=default_op)
        )
    # add the args that will be passed to the cle operation.
    for i, (name, type_, default) in enumerate(category_args):
        if i < len(category.default_values):
            default = category.default_values[i]
        params.append(Parameter(name, k, annotation=type_, default=default))

    # add a viewer.  This allows our widget to know if it's in a viewer
    params.append(
        Parameter(VIEWER_PARAM, k, annotation="napari.viewer.Viewer", default=viewer)
    )
    result = Signature(params)
    #print("Signature", result)
    return result

def _get_operation_name_for_category_clicked(category: Category, search_string:str= None) -> str:
    """Create an inspect.Signature object representing a cle Category.

    The output of this function is the operation name for a clicked cataegory given a search string
    """
    # Add valid operations choices (will create the combo box)
    from .._categories import operations_in_menu
    choices = list(operations_in_menu(category, search_string))

    default_op = category.default_op
    if not any(default_op == op for op in choices):
        #print("Default-operation is not in list!")
        default_op = None

    if default_op is None:
        return choices[0]
        
    else:
        return default_op

def make_gui_for_category(category: Category, search_string:str = None, viewer: napari.Viewer = None, button_size=DEFAULT_BUTTON_SIZE, operation_name:str=None, autocall:bool=True) -> magicgui.widgets.FunctionGui[Layer]:
    """Generate a magicgui widget for a Category object

    Parameters
    ----------
    category : Category
        An instance of a _categories.Category. (holds information about the cle operations,
        input types, and arguments that the widget needs to represent.)

    Returns
    -------
    magicgui.widgets.FunctionGui
        A magicgui widget instance
    """
    widget = None
    def gui_function(**kwargs) -> Optional[Layer]:
        """A function that calls a cle operation `call_op` and shows the result.

        This is the function that will be called by our magicgui widget.
        We modify it's __signature__ below.
        """
        viewer = kwargs.pop(VIEWER_PARAM, None)
        inputs = []
        for i in [kwargs.pop(k) for k in list(kwargs) if k.startswith("input")]:
            if isinstance(i, napari.layers.Layer):
                inputs.append(i)
            else:
                from napari_workflows._workflow import _get_layer_from_data
                inputs.append(_get_layer_from_data(viewer, i))

        result_layer = None
        t_position = None
        if viewer is not None and len(viewer.dims.current_step) == 4:
            # in case we process a 4D-data set, we need read out the current timepoint
            # and consider it further down in call_op
            t_position = viewer.dims.current_step[0]

            currstep_event = viewer.dims.events.current_step

            def update(event):
                #currstep_event.disconnect(update)
                if result_layer is not None:
                    from napari_workflows import WorkflowManager
                    manager = WorkflowManager.install(viewer)
                    manager.invalidate([result_layer.name])
                #widget()

            if hasattr(widget, 'updater'):
                currstep_event.disconnect(widget.updater)

            widget.updater = update

            currstep_event.connect(update)

        # todo: deal with 5D and nD data
        op_name = kwargs.pop("op_name")
        try:
            result, used_args = call_op(op_name, inputs, t_position, viewer, **kwargs)
        except TypeError as e:
            result = None
            used_args = []
            import warnings
            warnings.warn("Operation failed. Please check input parameters and documentation.\n" + str(e))

        # add a help-button
        description = find_function(op_name).__doc__
        if description is not None:
            description = description.replace("\n    ", "\n").replace("\n", "<br/>")
            getattr(widget, OP_NAME_PARAM).native.setToolTip("<html>" + description + "</html>")

        if result is not None:
            from napari.layers._source import layer_source
            with layer_source(widget=widget):
                result_layer = _show_result(
                    result,
                    viewer,
                    name=LAYER_NAME_PREFIX + f"{op_name}",
                    layer_type=category.output,
                    op_id=id(gui_function),
                    cmap=category.color_map,
                    blending=category.blending,
                    scale=inputs[0].scale,
                )
            if result_layer is None:
                return None

            # notify workflow manager that something was created / updated
            try:
                from napari_workflows import WorkflowManager
                manager = WorkflowManager.install(viewer)

                # this step basically separates actual arguments from kwargs as this can cause 
                # conflicts when setting the workflow step. 
                signat = signature(find_function(op_name))
                only_args = [
                    arg for arg, param in zip(used_args,signat.parameters.values()) 
                    if param.default is param.empty
                ]
                determined_kwargs = {
                    name:value for (name,param),value in zip(signat.parameters.items(),used_args) 
                    if not (param.default is param.empty)
                }
                # debugging prints
                #print(f'only arguments: {only_args}')
                #print(f'det kwargs:     {determined_kwargs}')
                manager.update(result_layer, find_function(op_name), *only_args, **determined_kwargs)
                #print("notified", result_layer.name, find_function(op_name))
            except ImportError:
                pass # recording workflows in the WorkflowManager is a nice-to-have at the moment.

            def _on_layer_removed(event):
                layer = event.value
                if layer in inputs or layer is result_layer:
                    try:
                        viewer.window.remove_dock_widget(widget.native)
                    # TODO find more elegant or specific solution 
                    # because this could break some stuff
                    except:
                        pass

            viewer.layers.events.removed.connect(_on_layer_removed)

            return result_layer
        return None

    gui_function.__name__ = f'do_{category.name.lower().replace(" ", "_")}'
    gui_function.__signature__ = _generate_signature_for_category(category, search_string, viewer)

    # create the widget
    widget = magicgui(gui_function, auto_call=autocall)
    widget.native.setMinimumWidth(100)
    modify_layout(widget.native, button_size=button_size)

    if operation_name == None:
        operation_name = _get_operation_name_for_category_clicked(category,search_string)
    
    # when the operation name changes, we want to update the argument labels
    # to be appropriate for the corresponding cle operation.
    op_name_widget = getattr(widget, OP_NAME_PARAM)
    op_name_widget.value = operation_name

    @op_name_widget.changed.connect
    def update_positional_labels(*_: Any):
        func = find_function(op_name_widget.value)
        new_sig = signature(func)
        # get the names of positional parameters in the new operation
        param_names, numeric_param_names, bool_param_names, str_param_names, file_param_names = separate_argnames_by_type(
            new_sig.parameters.items())
        num_count = 0
        str_count = 0
        bool_count = 0
        file_count = 0

        # show needed elements and set right label
        n_params = len(param_names)
        for n, arg in enumerate(category_args):
            arg_gui_name = arg[0]
            arg_gui_type = arg[1]
            wdg = getattr(widget, arg_gui_name)
            if arg_gui_type == FloatRange:
                if num_count < len(numeric_param_names):
                    arg_func_name = numeric_param_names[num_count]
                    num_count = num_count + 1
                else:
                    wdg.hide()
                    continue
            elif arg_gui_type == BoolType:
                if bool_count < len(bool_param_names):
                    arg_func_name = bool_param_names[bool_count]
                    bool_count = bool_count + 1
                else:
                    wdg.hide()
                    continue
            elif arg_gui_type == StringType:
                if str_count < len(str_param_names):
                    arg_func_name = str_param_names[str_count]
                    str_count = str_count + 1
                else:
                    wdg.hide()
                    continue
            elif arg_gui_type == PathLikeType:
                if file_count < len(file_param_names):
                    arg_func_name = file_param_names[file_count]
                    file_count = file_count + 1
                else:
                    wdg.hide()
                    continue
            else:
                arg_func_name = "?"
                print("Unsupported type:", arg_gui_type)
                continue

            wdg.label = arg_func_name
            wdg.text = arg_func_name
            wdg.show()

    # run it once to update the labels
    update_positional_labels()

    return widget


def modify_layout(widget, button_size = 32):
    QTimer.singleShot(100, partial(_modify_layout, widget, button_size))

def _modify_layout(widget, button_size=32):
    try:
        if hasattr(widget, "layout"):
            layout = widget.layout()
            if layout is not None:
                if isinstance(widget, QWidget):
                    layout.setContentsMargins(5, 1, 5, 1)
                    layout.setSpacing(0)

                for i in range(layout.count()):
                    w = layout.itemAt(i)
                    if isinstance(w.widget(), QDoubleSpinBox):
                        w.widget().setStyleSheet("""
                           QDoubleSpinBox::down-button{
                               width: {button_size};
                               height: {button_size}
                           }
                           QDoubleSpinBox::up-button{
                               height: {button_size};
                               width: {button_size};
                           }""".replace("{button_size}", str(button_size)))
                        w.widget().setMinimumHeight(button_size)
                    if isinstance(w.widget(), QComboBox):
                        w.widget().setMinimumHeight(button_size)
                        w.widget().view().setWordWrap(True)
                    modify_layout(w.widget(), button_size=button_size)
    except:
        pass

def separate_argnames_by_type(items):
    param_names = [
        name
        for name, param in items
        if param.annotation in {int, str, float, bool, PathLike}
    ]
    numeric_param_names = [
        name
        for name, param in items
        if param.annotation in {int, float}
    ]
    bool_param_names = [
        name
        for name, param in items
        if param.annotation in {bool}
    ]
    str_param_names = [
        name
        for name, param in items
        if param.annotation in {str}
    ]
    file_param_names = [
        name
        for name, param in items
        if param.annotation in {PathLike}
    ]
    return param_names, numeric_param_names, bool_param_names, str_param_names, file_param_names



from __future__ import annotations

from pathlib import Path
from typing import Callable
from warnings import warn
from qtpy.QtWidgets import QFileDialog, QLineEdit, QVBoxLayout, QHBoxLayout, QWidget, QMenu, QLabel
from qtpy.QtGui import QCursor
from typing import Union
from .._categories import CATEGORIES, Category, filter_categories, find_function, get_category_of_function
from ._button_grid import ButtonGrid, _get_highlight_brush, _get_background_brush
from ._category_widget import make_gui_for_category
from napari.viewer import Viewer

class Assistant(QWidget):
    """The main Assistant widget.

    The widget holds buttons with icons to create widgets for the various
    cel operation categories.  It tracks which layers are connected to which
    widgets, and can export the state of the task graph to a dask graph
    or to jython code.

    Parameters
    ----------
    napari_viewer : Viewer
        This viewer instance will be provided by napari when it gets added
        as a plugin dock widget.
    """

    def __init__(self, napari_viewer: Viewer):

        super().__init__()
        self._viewer = napari_viewer
        napari_viewer.layers.events.removed.connect(self._on_layer_removed)
        napari_viewer.layers.selection.events.changed.connect(self._on_selection)
        self._layers = {}

        # visualize intermediate results human-readable from top-left to bottom-right
        self._viewer.grid.stride = -1

        CATEGORIES["Generate code..."] = self._code_menu
        CATEGORIES["Save and load workflows"] = self._workflow_menu
        CATEGORIES["Undo"] = self.undo_action
        # CATEGORIES["Redo"] = self.redo_action

        CATEGORIES["Search napari hub"] = self.search_napari_hub
        CATEGORIES["Search image.sc"] = self.search_image_sc
        CATEGORIES["Search BIII"] = self.search_biii

        # build GUI
        self.icon_grid = ButtonGrid(self)
        self.icon_grid.addItems(CATEGORIES)
        self.icon_grid.itemClicked.connect(self._on_item_clicked)

        self.seach_field = QLineEdit("")
        self.seach_field.setPlaceholderText("Enter operation or plugin name to search")

        def text_changed(*args, **kwargs):
            search_string = self.seach_field.text().lower()
            self.icon_grid.clear()
            self.icon_grid.addItems(filter_categories(search_string))

        self.seach_field.textChanged.connect(text_changed)
        text_changed()

        # create menu
        self.actions = [
            ("Export Python script to file", self.to_python),
            ("Export Jupyter Notebook", self.to_notebook),
            ("Export Jupyter Notebook using Napari", self.to_notebook_using_napari),
            ("Copy Python code to clipboard", self.to_clipboard),
        ]

        # create workflow menu
        self.workflow_actions = [
            ("Export workflow to file", self.to_file),
            ("Load workflow from file", self.load_workflow),
        ]

        # add Send to script editor menu in case it's installed
        try:
            import napari_script_editor
            self.actions.append(("Send to Script Editor", self.to_script_editor))
        except ImportError:
            pass
        # add plugin generator menu in case it's installed
        try:
            import napari_assistant_plugin_generator
            self.actions.append(("Generate Napari plugin", self.to_napari_plugin))
        except ImportError:
            pass

        self.setLayout(QVBoxLayout())
        search_and_help = QWidget()
        search_and_help.setLayout(QHBoxLayout())
        from ._button_grid import _get_icon
        help = QLabel("?")
        help.setToolTip(
            '<html>'
            'Use the search field on the left to enter a term describing the function you would like to apply to your image.\n'
            'Searching will limit the number of shown categories and listed operations.\n'
            '<br><br>The icons in the buttons below denote the processed image types:\n'
            '<br><img src="' + _get_icon("intensity_image") + '" width="24" heigth="24"> In <b>intensity images</b> the pixel value represents a measurement, e.g. of collected light during acquisition in a microscope.\n'
            '<br><img src="' + _get_icon("binary_image") + '" width="24" heigth="24"> In <b>binary images</b> pixels with value 0 mean there is no object present. All other pixels (typically value 1) represent any object.\n'
            '<br><img src="' + _get_icon("label_image") + '" width="24" heigth="24"> In <b>label images</b> the integer pixel intensity corresponds to the object identity. E.g. all pixels of object 2 have value 2.\n'
            '<br><img src="' + _get_icon("parametric_image") + '" width="24" heigth="24"> In <b>parametric images</b> the pixel value represents an object measurement. All pixels of an object can for example contain the same value, e.g. the objects circularity or area.\n'
            '<br><img src="' + _get_icon("mesh_image") + '" width="24" heigth="24"> In <b>mesh images</b> we can visualize connectivity between objects and distances as intensity along lines.\n'
            '<br><img src="' + _get_icon("any_image") + '" width="24" heigth="24"> This icon means one can use <b>any kind of image</b> for this operation.'
            '</html>'
        )
        help.setMaximumWidth(20)
        search_and_help.layout().addWidget(self.seach_field)
        search_and_help.layout().addWidget(help)

        self.layout().addWidget(search_and_help)
        self.layout().addWidget(self.icon_grid)

        self.layout().setContentsMargins(5, 5, 5, 5)
        self.setMinimumWidth(345)

        self._on_selection()

    def _code_menu(self):
        menu = QMenu(self)

        for name, cb in self.actions:
            submenu = menu.addAction(name)
            submenu.triggered.connect(cb)

        menu.move(QCursor.pos())
        menu.show()

    def _workflow_menu(self):
        menu = QMenu(self)

        for name, cb in self.workflow_actions:
            submenu = menu.addAction(name)
            submenu.triggered.connect(cb)

        menu.move(QCursor.pos())
        menu.show()

    def _on_selection(self):
        for layer, (dw, gui) in self._layers.items():
            if layer in self._viewer.layers.selection:
                dw.show()    
            else:
                if layer in self._viewer.layers:
                    dw.hide()
        self._highlight_next_steps()

        try:
            _modify_layer_gui(self._viewer)
        except:
            pass

    def _highlight_next_steps(self):
        def dock_widget_category(name):
            if dw.name in CATEGORIES:
                return CATEGORIES[dw.name]
            return get_category_of_function(func_name=dw.name)

        highlighted_categories = []
        for layer, (dw, gui) in self._layers.items():
            if layer in self._viewer.layers.selection:
                category = dock_widget_category(dw.name) 
                if category is not None:
                    highlighted_categories += category.next_step_suggestions

                
        for key in CATEGORIES:
            try:
                self.icon_grid.item_mapping[key].setBackground(
                    _get_background_brush()
                )
            except RuntimeError:
                continue
            if key in highlighted_categories:
                self.icon_grid.item_mapping[key].setBackground(
                    _get_highlight_brush()
                )
        

    def _on_active_layer_change(self, event):
        for layer, (dw, gui) in self._layers.items():
            dw.show() if event.value is layer else dw.hide()

    def _on_layer_removed(self, event):
        layer = event.value
        if layer in self._layers:
            dw = self._layers[layer][0]
            try:
                self._viewer.window.remove_dock_widget(dw)
            except KeyError:
                pass
            # remove layer from internal list
            self._layers.pop(layer)


    def _on_item_clicked(self, item):
        self._activate(CATEGORIES.get(item.text()))

    def _get_active_layer(self):
        return self._viewer.layers.selection.active

    def _activate(self, category = Union[Category, Callable]):
        if callable(category):
            category()
            return

        # get currently active layer (before adding dock widget)
        input_layer = self._get_active_layer()
        if not input_layer:
            warn("Please select a layer first")
            return

        # make a new widget
        gui = make_gui_for_category(category, self.seach_field.text(), self._viewer,autocall=category.auto_call)
        # prevent auto-call when adding to the viewer, to avoid double calls
        # do this here rather than widget creation for the sake of
        # non-Assistant-based widgets.
        gui._auto_call = False
        # add gui to the viewer
        dw = self._viewer.window.add_dock_widget(gui, area="right", name=category.name)
        # workaround for https://github.com/napari/napari/issues/4348
        dw._close_btn = False
        # make sure the originally active layer is the input
        try:
            gui.input0.value = input_layer
        except ValueError:
            pass # this happens if input0 should be labels but we provide an image
        # call the function widget &
        # track the association between the layer and the gui that generated it
        if category.output in ['image', 'labels']:
            layer = gui()
            if layer is not None:
                self._layers[layer] = (dw, gui)
        # optionally turn on auto_call, and make sure that if the input changes we update
        gui._auto_call = category.auto_call
        self._connect_to_all_layers()
        self._on_selection()
        return gui

    def _refesh_data(self, event):
        self._refresh(event.source)

    def _refresh(self, changed_layer):
        """Goes through all layers and refreshs those which have changed_layer as input

        Parameters
        ----------
        changed_layer
        """
        for layer, (dw, mgui) in self._layers.items():
            for w in mgui:
                if w.value == changed_layer:
                    #mgui()
                    from napari_workflows import WorkflowManager
                    manager = WorkflowManager.install(self._viewer)
                    manager.invalidate([changed_layer.name])

    def _connect_to_all_layers(self):
        """Attach an event listener to all layers that are currently open in napari
        """
        for layer in self._viewer.layers:
            layer.events.data.disconnect(self._refesh_data)
            layer.events.data.connect(self._refesh_data)

    def load_sample_data(self, fname="Lund_000500_resampled-cropped.tif"):
        data_dir = Path(__file__).parent.parent / "data"
        self._viewer.open(str(data_dir / fname))

    def _id_to_name(self, id, dict):
        if id not in dict.keys():
            new_name = "image" + str(len(dict.keys()))
            dict[id] = new_name
        return dict[id]

    def to_python(self, filename=None):
        if not filename:
            filename, _ = QFileDialog.getSaveFileName(self, "Save code as...", ".", "*.py")
        #return Pipeline.from_assistant(self).to_jython(filename)

        from napari_workflows import WorkflowManager
        manager = WorkflowManager.install(self._viewer)
        code = manager.to_python_code()

        if filename:
            filename = Path(filename).expanduser().resolve()
            filename.write_text(code)

    def to_notebook_using_napari(self, filename=None, execute=True):
        return self.to_notebook(filename, execute, True)

    def to_notebook(self, filename=None, execute=True, use_napari=False):
        if not filename:
            filename, _ = QFileDialog.getSaveFileName(self, "Save code as notebook...", ".", "*.ipynb")
        #return Pipeline.from_assistant(self).to_notebook(filename)

        from napari_workflows import WorkflowManager
        manager = WorkflowManager.install(self._viewer)
        code = manager.to_python_code(notebook=True, use_napari=use_napari)

        import jupytext

        # jython code is created in the jupytext light format
        # https://jupytext.readthedocs.io/en/latest/formats.html#the-light-format

        jt = jupytext.reads(code, fmt="py:light")
        nb = jupytext.writes(jt, fmt="ipynb")
        if filename:
            filename = Path(filename).expanduser().resolve()
            filename.write_text(nb)
            # could use a NamedTemporaryFile to run this even without write
            if execute:
                from subprocess import Popen
                from shutil import which

                executable = "jupyter-lab"
                if not which(executable):
                    executable = "jupyter-notebook"
                    if not which(executable):
                        raise RuntimeError("Cannot find jupyter-lab or jupyter-notebook executable. Please install it using pip install jupyterlab")

                try:
                    Popen([executable, "-y", str(filename)])
                except Exception as e:
                    warn(f"Failed to execute notebook: {e}")
        return nb

    def to_napari_plugin(self):
        from napari_assistant_plugin_generator import make_plugin
        self._viewer.window.add_dock_widget(make_plugin())

    def to_clipboard(self):
        import pyperclip

        from napari_workflows import WorkflowManager
        manager = WorkflowManager.install(self._viewer)
        pyperclip.copy(manager.to_python_code())

    def to_script_editor(self):
        import napari_script_editor
        editor = napari_script_editor.ScriptEditor.get_script_editor_from_viewer(self._viewer)

        from napari_workflows import WorkflowManager
        manager = WorkflowManager.install(self._viewer)
        editor.set_code(manager.to_python_code())

    def to_file(self, filename=None):
        from napari_workflows import WorkflowManager, _io_yaml_v1

        if not filename:
            filename, _ = QFileDialog.getSaveFileName(self, "Export workflow ...", ".", "*.yaml")

        # get the workflow, should one be installed
        workflow_manager = WorkflowManager.install(self._viewer)
        _io_yaml_v1.save_workflow(filename, workflow_manager.workflow)

    def load_workflow(self, filename=None):
        from napari_workflows import _io_yaml_v1
        from .. _workflow_io_utility import initialise_root_functions, load_remaining_workflow

        layer_names = [str(lay) for lay in self._viewer.layers]
        if not layer_names:
            warn("No images opened. Please open an image before loading the workflow!")
            return

        if not filename:
            filename, _ = QFileDialog.getOpenFileName(self, "Import workflow ...", ".", "*.yaml")
        self.workflow = _io_yaml_v1.load_workflow(filename)

        w_dw = initialise_root_functions(
            self.workflow, 
            self._viewer,
        )
        w_dw += load_remaining_workflow(
            self.workflow, 
            self._viewer,
        )

        for gui, dw in w_dw:
            # call the function widget &
            # track the association between the layer and the gui that generated it
            category = get_category_of_function(
                find_function(gui.op_name.current_choice)
            )
            if category.output in ['image', 'labels']:
                layer = gui()
                if layer is not None:
                    self._layers[layer] = (dw, gui)

        self._viewer.layers.select_previous()
        self._viewer.layers.select_next()

    def undo_action(self):
        if len(self._viewer.dims.current_step) > 3:
            raise NotImplementedError("Undo/redo is not supported for 4D data (yet).")

        from .._undo_redo import clear_and_load_workflow, _change_widget_parameters#
        from napari_workflows import WorkflowManager, Workflow
        # install the workflow manager and get the current workflow and controller
        manager: WorkflowManager = WorkflowManager.install(self._viewer)
        controller = manager.undo_redo_controller
        workflow: Workflow = manager.workflow
        # only reload if there is an undo to be performed
        if controller.undo_stack:

            # undo workflow step: workflow is now the undone workflow
            undo_wf: Workflow = controller.undo()
            controller.freeze_stacks = True
            if set(undo_wf._tasks.keys()) == set(workflow._tasks.keys()):
                widgets_dict = {
                    key:self._layers[self._viewer.layers[key]][1] 
                    for key in workflow._tasks.keys()
                }
                _change_widget_parameters(
                    manager_workflow=workflow,
                    updated_workflow=undo_wf,
                    widgets=widgets_dict)
            else:
                clear_and_load_workflow(
                    viewer=self._viewer,
                    manager_workflow=workflow,
                    workflow_to_load=undo_wf,
                    layer_list=self._layers
                )

            controller.freeze_stacks = False            

    def redo_action(self):
        if len(self._viewer.dims.current_step) > 3:
            raise NotImplementedError("Undo/redo is not supported for 4D data (yet).")

        from .._undo_redo import clear_and_load_workflow, _change_widget_parameters
        from napari_workflows import WorkflowManager, Workflow

        # install the workflow manager and get the current workflow and controller
        manager: WorkflowManager = WorkflowManager.install(self._viewer)
        controller = manager.undo_redo_controller
        workflow = manager.workflow
        # only reload if there is an undo to be performed
        if controller.redo_stack:

            # undo workflow step: workflow is now the undone workflow
            redo_wf: Workflow = controller.redo()
            controller.freeze_stacks = True
            if len(redo_wf._tasks.keys()) == len(workflow._tasks.keys()):
                widgets_dict = {
                    key:self._layers[self._viewer.layers[key]][1] 
                    for key in workflow._tasks.keys()
                }
                _change_widget_parameters(
                    manager_workflow=workflow,
                    updated_workflow=redo_wf,
                    widgets=widgets_dict)
            else:
                clear_and_load_workflow(
                    viewer=self._viewer,
                    manager_workflow = workflow,
                    workflow_to_load=redo_wf,
                    layer_list=self._layers
                )

            controller.freeze_stacks = False

    def search_napari_hub(self):
        print("Search napari hub")
        from urllib.parse import quote
        _open_url("https://www.napari-hub.org/?search=" + quote(self.seach_field.text()) + "&sort=relevance")

    def search_image_sc(self):
        print("Search image sc")
        from urllib.parse import quote
        _open_url("https://forum.image.sc/search?q=napari%20" + quote(self.seach_field.text()))

    def search_biii(self):
        print("Search biii")
        from urllib.parse import quote
        _open_url("https://biii.eu/search?search_api_fulltext=napari%20" + quote(self.seach_field.text()))


def _open_url(url):
    import webbrowser
    webbrowser.open(url, new=0, autoraise=True)

# see https://github.com/napari/napari/issues/5127
def _modify_layer_gui(viewer):
    import warnings
    with warnings.catch_warnings():
        warnings.simplefilter(action='ignore', category=FutureWarning)
        from napari._qt.layer_controls.qt_layer_controls_container import QtLayerControlsContainer
        from napari._qt.layer_controls.qt_image_controls import QtImageControls
        from napari._qt.layer_controls.qt_image_controls_base import range_to_decimals
        from qtpy.QtWidgets import QPushButton

        # search for GUI in napari viewer
        obj = viewer.window.qt_viewer.dockLayerControls
        for i in range(obj.layout().count()):
            if obj.layout().itemAt(i) is not None and obj.layout().itemAt(i).widget is not None:
                item = obj.layout().itemAt(i).widget()
                if isinstance(item, QtLayerControlsContainer):
                    layer_controls = item
                    if layer_controls is not None:
                        obj2 = layer_controls
                        for i in range(obj2.layout().count()):
                            if obj2.layout().itemAt(i) is not None:
                                item = obj2.layout().itemAt(i).widget()
                                if isinstance(item, QtImageControls):
                                    image_controls = item

                                    if image_controls is not None:
                                        # Modify GUI: Add a "Reset" button to the auto-contrast button bar
                                        def reset_display_range():
                                            layer = list(viewer.layers.selection)[0]
                                            layer.contrast_limits_range = [layer.data.min(), layer.data.max()]
                                            layer.contrast_limits = list(layer.contrast_limits_range)

                                            contrastLimitsSlider = image_controls.contrastLimitsSlider
                                            decimals = range_to_decimals(
                                                layer.contrast_limits_range, layer.dtype
                                            )
                                            contrastLimitsSlider.setRange(*layer.contrast_limits_range)
                                            contrastLimitsSlider.setSingleStep(10 ** -decimals)
                                            contrastLimitsSlider.setValue(layer.contrast_limits)

                                        autoscalebar_widget = image_controls.autoScaleBar
                                        if autoscalebar_widget is not None and not hasattr(autoscalebar_widget, "_reset_button"):
                                            autoscalebar_widget._reset_button = QPushButton(('Reset'))
                                            autoscalebar_widget._reset_button.clicked.connect(reset_display_range)
                                            autoscalebar_widget.layout().addWidget(autoscalebar_widget._reset_button)
                                            autoscalebar_widget._auto_btn.setText("Cont.")




from napari_assistant._workflow_io_utility import load_remaining_workflow, initialise_root_functions
from napari_workflows import WorkflowManager, Workflow
from napari import Viewer
from ._gui._category_widget import DEFAULT_BUTTON_SIZE

def delete_workflow_widgets_layers(viewer: Viewer) -> None:
    """
    Deletes all layers which were created by widgets and closes the widgets prior

    Parameters
    ----------
    viewer: napari.Viewer
    """
    # install the workflow manager and get the current workflow
    manager: WorkflowManager = WorkflowManager.install(viewer)
    workflow = manager.workflow
    
    #deleting the widgets based on the workflow
    dock_widgets = viewer.window._dock_widgets
    workflow_nodes = list(workflow._tasks.keys())
    for key, widget in list(dock_widgets.items()):
        if ('Result of ' + key in workflow_nodes) or ('Result of '+key[:-22] in workflow_nodes):
            viewer.window.remove_dock_widget(widget=widget)
    
    #deleting the layers if created by a widget
    layers = viewer.layers
    layer_names = [str(lay) for lay in layers]
    for layer_name in layer_names:
        if layers[layer_name].source.widget:
            layers.remove(layer_name)

def clear_and_load_workflow(
    viewer: Viewer,
    manager_workflow: Workflow, 
    workflow_to_load: Workflow,
    layer_list:list,
    button_size = DEFAULT_BUTTON_SIZE
) -> list:
    """
    Loads a workflow by only initializing operations where the workflow
    is different from the currently loaded workflow.
    """
    delete_workflow_widgets_layers(viewer)
    manager_workflow.clear()

    w_dw = initialise_root_functions(
        workflow_to_load, 
        viewer, 
        button_size=button_size,
    )
    w_dw += load_remaining_workflow(
        workflow_to_load, 
        viewer,
        button_size=button_size,
    )

    from ._categories import find_function, get_category_of_function
    for gui, dw in w_dw:
        # call the function widget and
        # track the association between the layer and the gui that generated it
        category = get_category_of_function(
            find_function(gui.op_name.current_choice)
        )
        if category.output in ['image', 'labels']:
            layer = gui()
            if layer is not None:
                layer_list[layer] = (dw, gui)

    viewer.layers.select_previous()
    viewer.layers.select_next()

# EXPERIMENTAL        
def _change_widget_parameters(
    manager_workflow: Workflow, 
    updated_workflow:Workflow,
    widgets: dict,
) -> None:
    """
    Go through two workflows and update widgets where the workflows are different
    """
    from ._workflow_io_utility import category_kwargs, kwargs_of_wf_step

    for key in manager_workflow._tasks.keys():
        if manager_workflow._tasks[key] != updated_workflow._tasks[key]:
            kwargs = kwargs_of_wf_step(updated_workflow,key)
            for param_name, value in kwargs.items():
                if value in manager_workflow.sources_of(key):
                    pass # change input image
        
            widget = widgets[key]

            # setting the right parameters
            cat_kwargs = category_kwargs(
                updated_workflow._tasks[key][0],
                kwargs,
            )
            for k,v in cat_kwargs.items():
                widget[k].value = v

        



from napari_plugin_engine import napari_hook_implementation

from ._gui import Assistant
from ._categories import attach_tooltips

@napari_hook_implementation
def napari_experimental_provide_dock_widget():
    attach_tooltips()
    return [Assistant]

@napari_hook_implementation
def napari_experimental_provide_function():
    return [_split_stack]



def _split_stack(viewer:"napari.Viewer", layer:"napari.layers.Layer", axis:int = 1):
    # This function can be removed once these issues are fixed:
    # https://github.com/napari/napari/pull/3981
    # https://github.com/napari/napari/issues/4885
    from napari.layers.utils import stack_utils
    ll = viewer.layers

    # copied from: https://github.com/napari/napari/blob/f6bdd6235da23d4d23239653b6e4cf02e7f6d4ac/napari/layers/_layer_actions.py#L38
    images = stack_utils.stack_to_images(layer, axis)
    ll.remove(layer)
    ll.extend(images)
    ll.selection = set(images)  # type: ignore

try:
    from napari_tools_menu import register_dock_widget, register_function
    register_dock_widget(Assistant, menu="Utilities > Assistant (na)")
    register_function(_split_stack, menu="Utilities > Split stack (na)")

except ImportError:
    pass

